#version 450
#extension GL_ARB_separate_shader_objects : enable

// inputs
struct shader_input_0
{
	float2 texcoord : TEXCOORD0;
};

// outputs
struct shader_output_0
{
	float4 colour : COLOUR0;
};

// uniforms
layout(binding = 0) uniform ViewUniforms
{
	vec4 dims;
} view;

layout(binding = 1) uniform FogParams
{
	mat4 inverseVp;
	vec4 falloffParams;
	vec4 colour;
} fog;

layout(binding = 0) uniform sampler2D src;
layout(binding = 1) uniform sampler2D depthBuffer;

void main()
{
	vec4 colour = texture(src, shader_input_0.texcoord);
	float depth = texture(depthBuffer, shader_input_0.texcoord).r;
	
	vec4 screen = vec4(shader_input_0.texcoord, depth, 1.0);
	screen.xy = screen.xy * 2.0 - 1.0;
		
	vec4 world = fog.inverseVp * screen;
	world = world / world.w;
	
	float falloff = 0.0;
	
	if(depth < 1.0)
	{
		vec2 edgeDistance = abs(vec2(world.x, world.z)) - vec2(fog.falloffParams.x, fog.falloffParams.z) * 0.5;
		falloff = max(edgeDistance.x * 0.125, edgeDistance.y * 0.125);
		falloff = smoothstep(0.75, 0.875, falloff);
		falloff = falloff * mix(1.0, 0.0, smoothstep(0.8, 1.0, world.y / fog.falloffParams.y));	
	}
	else
	{
		vec4 camPos = fog.inverseVp * vec4(0.0, 0.0, 0.0, 1.0);
		camPos = camPos / camPos.w;
		camPos.z = camPos.z - 0.1;
		
		vec3 dir = normalize(world.xyz - camPos.xyz);
		
		vec3 topCentre = vec3(0.0, fog.falloffParams.y, 0.0);
		vec3 topNormal = vec3(0.0, 1.0, 0.0);
		
		vec3 centres[4] = 
		{
			vec3(0.0, 0.0, fog.falloffParams.z * 0.5),
			vec3(0.0, 0.0, fog.falloffParams.z * -0.5),
			vec3(fog.falloffParams.x * 0.5, 0.0, 0.0),
			vec3(fog.falloffParams.x * -0.5, 0.0, 0.0)
		};
		
		vec3 normals[4] =
		{
			vec3( 0.0, 0.0, -1.0),
			vec3( 0.0, 0.0,  1.0),
			vec3(-1.0, 0.0, 0.0),
			vec3( 1.0, 0.0, 0.0)
		};
		
		float closestT = 100000000.0;
		vec4 pos = vec4(0.0);
		for(int i = 0; i < 4; ++i)
		{
			// check for intersection with the world edge plane
			float denom = dot(normals[i], dir);
			if(abs(denom) > 0.0001)
			{
				float t = dot(centres[i] - camPos.xyz, normals[i]) / denom;
				if(t > 0.0 && t < closestT)
				{
					// check if top bounding plane intersection is in front or behind world edge
					vec3 origin = camPos.xyz + dir * t;
					if(origin.y > fog.falloffParams.y)
					{				
						denom = dot(topNormal, dir);
						if(abs(denom) > 0.0001)
						{
							float topT = dot(topCentre - origin, topNormal) / denom;
							if(topT >= 0.0)
							{
								// top plane of fog hit								
								world.xyz = origin + dir * topT;
								
								// falloff the fog as the angle between the normal and the dir approaches 90 degrees
								falloff = smoothstep(0.0, 0.05, abs(denom));
							}
							else
							{
								// sky hit
								falloff = 0.0;
							}
						}
					}
					else
					{
						world.xyz = origin;
						falloff = 1.0;
					}
					
					closestT = t;
				}
			}
		}
		
		
	}
	
	shader_output_0.colour.xyz = mix(colour.xyz, fog.colour.xyz, falloff);
	shader_output_0.colour.w = 1.0f;
}